-- implemenation of pattern matching for Euphoria
-- v0.3
-- email: dcuny@lanset.com
-- (c) 2000 David Cuny


constant
    FALSE   = 0,
    TRUE    = not FALSE

constant
    AND             = 1,        -- tests must all be true
    OR              = 2,        -- one test must be true
    IS              = 3,        -- must match the whole pattern
    IS_NOT          = 4,        -- must not match the whole pattern
    CHAR            = 5,        -- must match char
    OPEN            = 6,        -- open parenthesis
    CLOSE           = 7,        -- close parenthesis
    START           = 8,        -- start of line
    END             = 9,        -- end of line
    WORD_BOUND      = 10,       -- word boundary
    NON_WORD_BOUND  = 11        -- non-word boundary

constant
    MORE        = -1        -- for specifying open-ended range

constant
    CODE        = 1,    -- rule code
    MIN         = 2,    -- minimum occurances
    MAX         = 3,    -- maximum occurances
    RANGE       = 4     -- characters to match

    
sequence parseString
sequence targetString
integer targetAt
sequence argStack

-- globals
global sequence arg         -- bindings
global integer args         -- count of bindings


-------------------------------------------------------------------------------
function isWordChar( integer char )

    -- true if char is word char
    if    char >= 'A' and char <= 'Z' then  return TRUE
    elsif char >= 'a' and char <= 'z' then  return TRUE
    elsif char >= '0' and char <= '9' then  return TRUE
    elsif char = '_' then                   return TRUE
    end if
    
    return FALSE
    
end function


-------------------------------------------------------------------------------
function applyPattern( sequence rule )

    -- execute the grep rules against the string
    integer count, isRule, result, rollback, char, flag, index, at
    sequence string

    -- AND
    if rule[CODE] = AND then  
        -- test form:
        --      { AND, test, test ... test n }
    
        -- check each rule
        for i = 2 to length( rule ) do
            -- rule failed?
            if not applyPattern( rule[i] ) then
                return FALSE
            end if
        end for
        
        -- all the rules succeeded
        return TRUE
    
    -- OR
    elsif rule[CODE] = OR then
        -- test form:
        --      { OR, test, test ... test n }
    
        -- check each rule
        for i = 2 to length( rule ) do     
            
            -- set rollback
            rollback = targetAt
            
            -- rule succeeded?
            if applyPattern( rule[i] ) then
                return TRUE
            end if
            
            -- rollback
            targetAt = rollback
            
        end for
        
        -- all the rules failed
        return FALSE

    -- IS or IS_NOT    
    elsif rule[CODE] = IS
    or    rule[CODE] = IS_NOT then
        -- test form:
        --      { IS,     low, high, range string }
        --      { IS_NOT, low, high, range string }
        
        -- set the count to zero
        count = 0
    
        -- get is rule - either IS or IS_NOT
        isRule = rule[CODE]
    
        -- get the range string
        string = rule[RANGE]
        
        -- try to apply the rule
        while TRUE do      
        
            -- mark undo point
            rollback = targetAt
        
            -- get char from target string
            if targetAt > length( targetString ) then
                -- past length of target             
                char = -1
            else
                char = targetString[targetAt]
            end if

            -- set result of test to false
            flag = FALSE
        
            -- does it match one of the ranges in the string?
            for i = 1 to length( string ) by 2 do
                
                -- test
                result = ( char >= string[i] and char <= string[i+1] )
                          
                if ( isRule = IS and result )
                or ( isRule = IS_NOT and not result ) then
                
                    -- set test flag
                    flag = TRUE
                    
                    -- exit loop
                    exit
                    
                end if
                
            end for 
            
            -- failed?
            if flag = FALSE then    
                -- rollback to last good position
                targetAt = rollback              
                
                -- exit
                exit
                
            else
                -- increment the count
                count += 1  
            
                -- move target position
                targetAt += 1
            
            end if
    
            -- reached upper limit?
            if count >= rule[MAX]
            and rule[MAX] != MORE then
                -- leave the loop
                exit
            end if

            -- reached end of string?
            if char = -1 then
                -- leave loop
                exit
            end if  
            
        end while

        -- evaluate the success of the match
        if rule[MIN] = 0 then
            -- zero or matches
            return TRUE

        elsif count < rule[MIN] then
            -- didn't meet minimum requirement
            return FALSE
                       
        elsif rule[MAX] != MORE
        and   count < rule[MAX] then
            -- didn't meet minimum requirement
            return FALSE
                       
        else
            -- matched enough
            return TRUE
            
        end if
        
    -- CHAR
    elsif rule[CODE] = CHAR then
        -- test form:
        --      { CHAR, char }

        -- past end of target?
        if targetAt > length( targetString ) then
            return FALSE
        end if

        -- match?
        if targetString[targetAt] = rule[2] then

            -- move ahead
            targetAt += 1
            
            -- return true
            return TRUE
            
        end if

        -- not matched
        return FALSE

    elsif rule[CODE] = OPEN then
        -- beginning of parenthesis
        -- push count and start point onto stack
        
        args += 1
        argStack = prepend( argStack, { args, targetAt} )

        -- by definition, true
        return TRUE
             
             
    elsif rule[CODE] = CLOSE then
        -- end of parenthesis
        -- pop count and start point from stack

        -- the index of the item        
        index = argStack[1][1]
        
        -- start position of string
        at = argStack[1][2]
        
        -- pop the stack 
        argStack = argStack[2..length(argStack)]

        -- check if pointer is past end of string
        if targetAt > length( targetString ) then 
            -- limit to one char beyond
            targetAt = length( targetString ) + 1
        end if

        -- assign the string
        arg[index] = targetString[at..targetAt-1]

        -- by definition, true
        return TRUE

    -- start of line?
    elsif rule[CODE] = START then
        return (targetAt = 1)
                        
    -- end of line?
    elsif rule[CODE] = END then
        return (targetAt > length( targetString ))

    -- word boundary?
    elsif rule[CODE] = WORD_BOUND then
        -- is there a prior character?
        if targetAt = 1 then
            -- true if current char is legal word char
            return isWordChar( targetString[targetAt] )
            
        else  
            -- was prior char part of a word?
            if isWordChar( targetString[targetAt-1] ) then
                return FALSE
            else
                -- true if current char is a word char
                return isWordChar( targetString[targetAt] )
            end if
            
        end if
                        
    -- non-word boundary?
    elsif rule[CODE] = NON_WORD_BOUND then
        -- essentially a negation of WORD_BOUND test
    
        -- is there a prior character?
        if targetAt = 1 then
            -- true if current char not is legal word char
            return not isWordChar( targetString[targetAt] )
            
        else  
            -- was prior char not part of a word?
            if not isWordChar( targetString[targetAt-1] ) then
                return FALSE
            else
                -- true if current char is not a word char
                return not isWordChar( targetString[targetAt] )
            end if
            
        end if
    

                        
    else
        -- error!     
        printf( 1, "Error in applyPattern: unknown code %d.\n", {rule[CODE]} )
        abort(0)
        
    end if  
    
end function


-------------------------------------------------------------------------------
function leftIs( sequence leftSide )

    -- return true if left hand of string matches leftSide
    -- removes characters from string on match

    integer size

    -- get size of string to match    
    size = length( leftSide )

    -- match string larger than target?
    if length( parseString ) < size then
        -- no match
        return FALSE                                    
        
    elsif equal( leftSide, parseString[1..size] ) then
        -- remove string
        parseString = parseString[size+1..length(parseString)]
        
        -- match
        return TRUE
        
    else
        -- no match
        return FALSE
        
    end if             
    
end function


-------------------------------------------------------------------------------
procedure removeChar()
    -- remove leftmost character of parseString
    parseString = parseString[2..length(parseString)]
end procedure


-------------------------------------------------------------------------------
function parseModifier( sequence code )

    -- parse modifier, if there is one
    
    integer number

    -- ?: 0 or 1 of the pattern
    if leftIs( "?" ) then
        code[MIN] = 0
        code[MAX] = 1

    -- *: 0 or many                
    elsif leftIs( "*" ) then
        code[MIN] = 0
        code[MAX] = MORE
                        
    -- +: 1 or many                
    elsif leftIs( "+" ) then
        code[MIN] = 1
        code[MAX] = MORE

    -- {m,n}
    elsif leftIs( "{" ) then
    
        -- parse a number, up to comma
        number = 0
        while TRUE do
        
            if leftIs( "," ) then     
            
                -- is pattern {n,}
                if leftIs( "}" ) then
                    -- code as {n,MORE}
                    code[MIN] = number
                    code[MAX] = MORE
                
                -- pattern is {n,m}
                else       
                    -- store min
                    code[MIN] = number
                    
                    -- parse second number, up to }
                    number = 0
                    while not leftIs( "}" ) do
                    
                        -- accumulate a digit
                        number = (number * 10) + ( parseString[1] - '0' )
                        
                        -- move ahead
                        removeChar()
                        
                    end while
                    code[MAX] = number
                
                    -- all done
                    exit
            
                end if
            
            -- pattern is {n}                        
            elsif leftIs( "}" ) then
                -- missing second number; needs exact match
                code[MIN] = number
                code[MAX] = number
                
                -- all done
                exit

            -- still reading first number following {...                        
            else

                -- accumulate a digit
                number = (number * 10) + ( parseString[1] - '0' )
                
                -- move ahead
                removeChar()
        
            end if
            
        end while

    end if

    return code
        
end function



-------------------------------------------------------------------------------
function parsePatternBlock()

    integer char
    sequence orList, rules, rule

    -- if or is encountered, rules are stored in here
    orList = {}
    
    -- this is the current list of accumulated rules
    rules = { AND }
    
    while TRUE do    

        -- clear rule
        rule = {}
    
        if length( parseString ) = 0
        or parseString[1] = ' '
        or parseString[1] = ')' then

            -- do these rules need to be OR'd together?
            if length( orList ) then
                -- add last rules to list
                orList = append( orList, rules )
                
                -- return orList
                return orList
                
            else
            
                -- return accumulated rules
                return rules
                
            end if

        elsif leftIs( "|" ) then
        
            -- first or encountered?
            if length( orList ) = 0 then
                -- place accumulated rules in or list
                orList = { OR, rules }
        
            else
                -- add rules to or list
                orList = append( orList, rules )        
                
            end if
                             
            -- clear rules
            rules = { AND }


        -- start of line
        elsif leftIs( "^" ) then
            rule = { START }

        -- start of line
        elsif leftIs( "$" ) then
            rule = { END }

        -- anything but newline        
        elsif leftIs( "." ) then
        
            -- eat one non-newline
            rule = { IS_NOT, 1, 1, "\n\n" }

            -- allow modifier
            rule = parseModifier( rule )

        -- set of characters            
        elsif leftIs( "[" ) then
            -- expression matches any single character not in set
            
            if leftIs( "^" ) then
            
                -- rule is IS
                rule = { IS_NOT, 1, 1, "" }
    
            else
            
                -- rule is IS
                rule = { IS, 1, 1, "" }
                
            end if

            -- parse until right brace
            while not leftIs( "]" ) do

                -- badly formed?
                if length( parseString ) = 0 then
                    puts( 1, "Missing ']'.\n" )
                    abort(0)
                end if
                
                -- get low range value
                char = parseString[1]
                rule[RANGE] &= char
                removeChar()
                
                -- hyphen?
                if leftIs( "-" ) then
                    -- get high range
                    rule[RANGE] &= parseString[1]
                    
                    -- remove high range
                    removeChar()
                    
                else
                    -- same as low range
                    rule[RANGE] &= char
                end if

            end while    
            
            -- allow modifier
            rule = parseModifier( rule )

        elsif leftIs( "(" ) then
            -- add open paren
            rules = append( rules, {OPEN} )
            
            -- add rules
            rules = append( rules, parsePatternBlock() ) 

        elsif leftIs( "/b" ) then
            -- falls on word boundary
            rule = { WORD_BOUND }
                         
        elsif leftIs( "/B" ) then
            -- falls on non-word boundary
            rule = { NON_WORD_BOUND }
                         

        elsif leftIs( "/d" ) then
            -- matches a digit, same as [0-9]
            rule = { IS, 1, 1, "09" }
            
            -- allow modifier
            rule = parseModifier( rule )
                        

        elsif leftIs( "/D" ) then
            -- matches a non-digit, same as [^0-9]
            rule = { IS_NOT, 1, 1, "09" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/w" ) then
            -- matches word char [a-zA-Z0-9_]
            rule = { IS, 1, 1, "azAZ09__" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/W" ) then
            -- matches non-word char [^a-zA-Z0-9_]
            rule = { IS_NOT, 1, 1, "azAZ09__" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/s" ) then
            -- matches whitespace char (space, tab, newline...)
            rule = { IS, 1, 1, "  \n\n\r\r\t\t" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/S" ) then
            -- matches non whitespace char (space, tab, newline...)
            rule = { IS_NOT, 1, 1, "  \n\n\r\r\t\t" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/n" ) then
            -- matches newline
            rule = { IS, 1, 1, "\n\n" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/r" ) then
            -- matches return          
            rule = { IS, 1, 1, "\r\r" }
                    
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/t" ) then
            -- matches tab
            rule = { IS, 1, 1, "\t\t" }
            
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/f" ) then
            -- matches formfeed  
            puts( 1, "/f not implemented yet.\n" )
            abort(0)
            

        elsif leftIs( "/b" ) then
            -- matches backspace (inside [] only)
            puts( 1, "/b not implemented yet.\n" )
            abort(0)
            

        elsif leftIs( "/0" ) then
            -- matches null character    
            rule = { IS, 1, 1, {0,0} }        
    
            -- allow modifier
            rule = parseModifier( rule )
            

        elsif leftIs( "/x" ) then
            -- matches hexidecimal value /xNNN
            puts( 1, "/x not implemented yet.\n" )
            abort(0)
            

        elsif leftIs( "/c" ) then
            -- matches control character /cX
            puts( 1, "/c not implemented yet.\n" )
            abort(0)
            

        elsif leftIs( "/" ) then
            -- matches following literal character
            rule = { IS, 1, 1, parseString[1] & parseString[1] }
            removeChar()
            
            -- allow modifier
            rule = parseModifier( rule )                     

        else
            -- character literal
            rule = { CHAR, parseString[1] }
            removeChar()
            
        end if

        -- rule created?
        if length( rule ) then
            rules = append( rules, rule )
        end if

        
    end while

    
end function


-------------------------------------------------------------------------------
global function pattern( sequence pattern, sequence target )

    -- top level pattern matching 
    sequence rules
    
    -- rules are AND'ed together
    rules = { AND }

    -- store pattern in parseString
    parseString = pattern

    -- parse to end of string
    while length( parseString ) do
        -- parse next set of rules
        rules = append( rules, parsePatternBlock() )
        
        -- white space encountered?
        if leftIs( " " ) then
        
            -- rule to eat spaces
            rules = append( rules, { IS, 1, MORE, "  " } )
            
            -- remove white space from parseString
            while leftIs( " " ) do
                -- leftIs removes the matching char
            end while

        elsif leftIs( ")" ) then            
        
            -- add close paren to rules
            rules = append( rules, {CLOSE} )
            
        end if

        
    end while
    
    -- clear the args
    arg = repeat( {}, 32 )                       

    for i = 1 to length( target )-1 do
    
        -- initialize the arg stack
        args = 0
        argStack = {}               

        -- initialize the target string to parse
        targetAt = 1
        targetString = target[i..length(target)]

        -- apply the parsed pattern to the target                       
        if applyPattern( rules ) then
            return TRUE
        end if
        
    end for   
    
    -- no matches                   
    return FALSE
              
end function

